#ifndef Q_MATH_LSL
#define Q_MATH_LSL

// should make most of these #defines - to take advantage of pseudo polymorphism

#define qMix(min, max, value) (((max) - (min)) * (value) + (min))

float qMin(float a, float b)
{
  if (a < b)
    return a;
  else
    return b;
}

float qMax(float a, float b)
{
  if (a > b)
    return a;
  else
    return b;
}


float qClamp(float min, float max, float value)
{
  if (value < min)
    return min;
  if (value > max)
    return max;

  return value;
}



float qScale(float origMin, float origMax, 
	     float newMin, float newMax, 
	     float value)
{
  float x = (value - origMin)/(origMax - origMin);
  
  return newMin + x * (newMax - newMin);
}



float qSmoothstep (float edge0, float edge1, float x)
{
   if (x < edge0)
      return 0; 

   if (x >= edge1)
      return 1; 

   x = (x - edge0) / (edge1 - edge0); 

   return x * x * (3.0 - 2.0 * x);
}


float qMod(float value, float divisor)
{
  integer m = llFloor(value / divisor);

  return value - m * divisor;
}

float qRandom(float min, float max)
{
  return llFrand(max - min) + min;
}




// computes bspline index - for either closed (periodic) or open bsplines.
integer qSplineIndex(integer index, integer length, integer closed)
{
  if (closed)
    return (integer) qMod(index, length);

  if (index > length - 1)
    return length - 1;

  if (index < 0)
    return 0;

  return index;
}

// vector spline function - takes a list of
// vectors and creates a spline over
// range [0..llGetListLength(vectors)] (uniform


vector qSplineV(list vectors, float value, integer closed)
{
  float length = llGetListLength(vectors);

  float floor   = (float)llFloor(value);
  float remains = value - floor;
  
  integer i1 = qSplineIndex((integer)(floor - 1.0), (integer)length, closed);
  integer i2 = qSplineIndex((integer)(floor + 0.0), (integer)length, closed);
  integer i3 = qSplineIndex((integer)(floor + 1.0), (integer)length, closed);
  integer i4 = qSplineIndex((integer)(floor + 2.0), (integer)length, closed);

  vector v1 = llList2Vector(vectors, i1);
  vector v2 = llList2Vector(vectors, i2);
  vector v3 = llList2Vector(vectors, i3);
  vector v4 = llList2Vector(vectors, i4);

  // weights derived from bspline matrix - [Foley1990, p. 493]
  float t  = remains;
  float t2 = t * remains;
  float t3 = t2 * remains;

  float w1 = -t3 + (3.0 * t2) - (3.0 * t) + 1.0;
  float w2 = (3.0 * t3) - (6.0 * t2) + 4.0;
  float w3 = (-3.0 * t3) + (3.0 * t2) + (3.0 * t) + 1.0;
  float w4 = t3;


  vector p = ( (w1 * v1) + (w2 * v2) + (w3 * v3) + (w4 * v4) ) / 6.0;

  return p;
}


// calculate approximate length of spline by stepping in segments.
// (you know, i bet this has a closed form.  too bad i'm too lazy/stupid to derive it.)
float qSplineVLength(list vectors, float v1, float v2, integer segments, integer closed)
{
  float length = 0;

  float segmentDelta = (v2 - v1) / segments;

  integer i;
  for (i = 0; i < segments; i++)
    {
      float s1 = v1 + segmentDelta * i;
      float s2 = s1 + segmentDelta;

      vector p1 = qSplineV(vectors, s1, closed);
      vector p2 = qSplineV(vectors, s2, closed);

      length += llVecDist(p1, p2);
    }

  return length;
}





rotation qRotationScale(float a, rotation r)
{
  rotation s;
  
  s.x = r.x * a;
  s.y = r.y * a;
  s.z = r.z * a;
  s.s = r.s * a;

  return s;
}



rotation qRotationMix(rotation a, rotation b, float t)
{
  float o = llAngleBetween(a, b) / 2.0;

  if (llFabs(llSin(o)) < 0.001)
    return a;

  rotation c = (qRotationScale(llSin((1.0 - t) * o), a) +
		qRotationScale(llSin(t * o), b));

  c = qRotationScale(1.0 / llSin(o), c);
			      
  return c;
}


string qFloat2String(float value, integer precision)
{
  integer base    = llFloor(value);
  float   remainder = value - base;

  string valueStr = (string) base;

  if (remainder != 0)
    {
      string remainderStr = (string) remainder;
      valueStr += llGetSubString(remainderStr, 1, precision + 1);
    }

  return valueStr;
}


string qVector2String(vector value, integer precision)
{

  return ("< " + 
	  qFloat2String(value.x, precision) + ", " +
	  qFloat2String(value.y, precision) + ", " +
	  qFloat2String(value.z, precision) + " >");
	  
}

#endif

